import re
import os


def quote_ifstr(s):
    return f"'{s}'" if isinstance(s, str) else s


def make_str_valid_varname(s):
    # remove invalid characters (except spaces in-between)
    s = re.sub(r"[^0-9a-zA-Z_\s]", " ", s).strip()

    # replace spaces by underscores (instead of dropping spaces) for readability
    s = re.sub(r"\s+", "_", s)

    # if first char is not letter or underscore then add underscore to make it valid
    if not re.match("^[a-zA-Z_]", s):
        s = "_" + s

    return s


def try2eval(val, globals={}, locals={}):
    try:
        return eval(val, globals, locals)
    except Exception:
        return val


def isiterable(x):
    """Returns True if `x` is iterable that is not a string or dict and is not null"""
    from spyql.nulltype import Null
    from collections.abc import Iterable

    return (
        isinstance(x, Iterable)
        and x is not Null
        and not isinstance(x, str)
        and not isinstance(x, dict)
    )


def is_row_collapsable(row, colnames):
    """
    Returns True if `row` only has a single column of type dict with a default name.
    In this case, a row (of type dict) can take the value of the first column.
    """
    return (
        len(row) == 1
        and isinstance(row[0], dict)
        and colnames[0] in {"col1", "json", "row"}
    )


def join_paths(x, *args):
    """convienience function for os.path.join"""
    return os.path.join(x, *args)



import logging

logging.basicConfig(level=logging.INFO)


from click.testing import CliRunner
import spyql.cli
import spyql.log
from spyql.writer import SpyWriter
from spyql.processor import SpyProcessor
from spyql.nulltype import NULL
from spyql.qdict import qdict
from tabulate import tabulate
import json
import csv
import io
import sqlite3
import math
from functools import reduce
import sys


# --------  AUX FUNCTIONS  --------
def json_output(out):
    return [
        json.loads(line, object_hook=lambda x: qdict(x, dirty=False))
        for line in out.splitlines()
    ]


def txt_output(out, has_header=False):
    if has_header and out.count("\n") == 1:
        return []  # special case when outputs is the header row (output has no data)
    return out.splitlines()


def list_of_struct2pretty(rows):
    if not rows:
        return []
    vals = [  # NULLs should be replaced by Nones before calling tabulate
        {key: (None if val is NULL else val) for (key, val) in row.items()}
        for row in rows
    ]
    return (tabulate(vals, headers="keys", tablefmt="simple") + "\n").splitlines()


def list_of_struct2csv(vals):
    if not vals:
        return []
    out_str = io.StringIO()
    writer = csv.DictWriter(out_str, vals[0].keys())
    writer.writeheader()
    for val in vals:
        writer.writerow(val)
    return out_str.getvalue().splitlines()


def list_of_struct2py(vals):
    if not vals:
        return []
    header = [str(list(vals[0].keys()))]
    rows = [str(tuple(val.values())) for val in vals]
    return header + rows


def spy2py(lines):
    return [str(SpyProcessor.unpack_line(line)) for line in lines]


def run_query(query, data, **kw_options):
    bk = sys.stdin
    sys.stdin = io.StringIO(data)
    res = spyql.query.Query(query, **kw_options)()
    sys.stdin = bk
    return res


def run_cli(query="", options=[], data=None, runner=CliRunner(), exception=True):
    res = runner.invoke(spyql.cli.main, options + [query], input=data)
    return res


def make_cli_options(kw_options):
    options = []
    for option, value in kw_options.items():
        if option.endswith("put_options"):
            options.extend(
                [
                    f"-{option[0].upper()}{opt}" + ("" if val is None else f"={val}")
                    for opt, val in value.items()
                ]
            )
        elif option == "unbuffered":
            if value:
                options.extend(["-u"])
        elif option == "warning_flag":
            options.extend([f"-W{value}"])
        elif option == "v":
            options.extend([f"-v{value}"])
        else:
            options.extend([f"--{option}"])
    return options


def eq_test_nrows(query, expectation, data=None, **kw_options):
    runner = CliRunner()
    spyql.log.user_info("Running query: {}".format(query))

    options = make_cli_options(kw_options)

    res = run_cli(query + " TO json", options, data, runner)
    assert json_output(res.output) == expectation
    assert res.exit_code == 0

    res = run_cli(query + " TO orjson", options, data, runner)
    assert json_output(res.output) == expectation
    assert res.exit_code == 0

    res = run_cli(query + " TO csv", options, data, runner)
    assert txt_output(res.output, True) == list_of_struct2csv(expectation)
    assert res.exit_code == 0

    res = run_cli(query + " TO spy", options, data, runner)
    assert spy2py(txt_output(res.output, True)) == list_of_struct2py(expectation)
    assert res.exit_code == 0

    res = run_cli(query + " TO pretty", options, data, runner)
    assert txt_output(res.output, True) == list_of_struct2pretty(expectation)
    assert res.exit_code == 0

    res = run_query(query + " TO memory", data, **kw_options)
    assert res == tuple(expectation)


def eq_test_1row(query, expectation, **kwargs):
    eq_test_nrows(query, [expectation], **kwargs)


def exception_test(query, anexception, **kw_options):
    res = run_cli(query, make_cli_options(kw_options))
    assert res.exit_code != 0
    assert isinstance(res.exception, anexception)


# --------  TESTS  --------
def test_basic():

    # single column
    # int
    eq_test_1row("SELECT 1", {"_1": 1})
    eq_test_1row("SELECT 1+2", {"_1_2": 3})

    # float
    eq_test_1row("SELECT 1.1", {"_1_1": 1.1})
    eq_test_1row("SELECT 1+0.2", {"_1_0_2": 1.2})

    # text
    eq_test_1row("SELECT '1'", {"_1": "1"})
    eq_test_1row("SELECT '1'+'2'", {"_1_2": "12"})

    # two columns with differemt data types
    eq_test_1row("SELECT '1', 2", {"_1": "1", "_2": 2})

    # alias
    eq_test_1row("SELECT '1' as a, 2 AS Ola", {"a": "1", "Ola": 2})

    # strings with commas and reserved keywords
    eq_test_1row(
        "SELECT 'isto, from you' as 'era uma vez', 2 AS Ola",
        {"era uma vez": "isto, from you", "Ola": 2},
    )

    # star over a literal
    eq_test_1row("SELECT * FROM 1", {"col1": 1})

    # star over a list
    eq_test_1row("SELECT * FROM [1]", {"col1": 1})
    # TODO: star over JSON, CSV, DIC?, DF? ZIPPED LIST ...

    # get all elements from a list/iterator
    eq_test_nrows("SELECT * FROM [1,2,3]", [{"col1": 1}, {"col1": 2}, {"col1": 3}])

    # where clause
    eq_test_nrows("SELECT * FROM [1,2,3] WHERE col1 >= 2", [{"col1": 2}, {"col1": 3}])

    # where filters out all rows
    eq_test_nrows("SELECT * FROM [1,2,3] WHERE col1 >= 10", [])

    # where + limit all
    eq_test_nrows(
        "SELECT * FROM [1,2,3] WHERE col1 >= 2 LIMIT ALL", [{"col1": 2}, {"col1": 3}]
    )

    # where + large limit
    eq_test_nrows(
        "SELECT * FROM [1,2,3] WHERE col1 >= 2 LIMIT 1000", [{"col1": 2}, {"col1": 3}]
    )

    # limit
    eq_test_nrows("SELECT * FROM [1,2,3] LIMIT 2", [{"col1": 1}, {"col1": 2}])

    # limit and offset 0
    eq_test_nrows("SELECT * FROM [1,2,3] LIMIT 2 OFFSET 0", [{"col1": 1}, {"col1": 2}])

    # limit and negative offset
    eq_test_nrows("SELECT * FROM [1,2,3] LIMIT 2 OFFSET -2", [{"col1": 1}, {"col1": 2}])

    # limit and offset
    eq_test_nrows(
        "SELECT * FROM [10,20,30,40,50,60] LIMIT 2 OFFSET 3",
        [{"col1": 40}, {"col1": 50}],
    )

    # offset only
    eq_test_nrows(
        "SELECT * FROM [10,20,30,40,50,60] OFFSET 2",
        [{"col1": 30}, {"col1": 40}, {"col1": 50}, {"col1": 60}],
    )

    # limit 0
    eq_test_nrows("SELECT * FROM [1,2,3] LIMIT 0", [])

    # negative limit
    eq_test_nrows("SELECT * FROM [1,2,3] LIMIT -10", [])

    # complex expressions with commas and different types of brackets
    eq_test_1row(
        "SELECT (col1 + 3) + ({'a': 1}).get('b', 6) + [10,20,30][(1+(3-2))-1] AS calc,"
        " 2 AS two FROM [1]",
        {"calc": 30, "two": 2},
    )

    # import
    eq_test_1row(
        "IMPORT sys SELECT sys.version_info.major AS major_ver", {"major_ver": 3}
    )

    eq_test_1row(
        "IMPORT numpy AS np, sys SELECT (np.array([1,2,3])+1).tolist() AS a",
        {"a": [2, 3, 4]},
    )


def test_orderby():
    # order by (1 col)
    eq_test_nrows(
        "SELECT * FROM [1,-2,3] ORDER BY 1", [{"col1": -2}, {"col1": 1}, {"col1": 3}]
    )
    eq_test_nrows(
        "SELECT * FROM [1,-2,3] ORDER BY 1 DESC",
        [{"col1": 3}, {"col1": 1}, {"col1": -2}],
    )
    eq_test_nrows(
        "SELECT abs(col1) as col1 FROM [1,-2,3] ORDER BY 1",
        [{"col1": 1}, {"col1": 2}, {"col1": 3}],
    )
    eq_test_nrows(
        "SELECT abs(col1) as col1 FROM [1,-2,3] ORDER BY 1 DESC",
        [{"col1": 3}, {"col1": 2}, {"col1": 1}],
    )
    eq_test_nrows(
        "SELECT col1 FROM [1,-2,3] ORDER BY col1",
        [{"col1": -2}, {"col1": 1}, {"col1": 3}],
    )
    eq_test_nrows(
        "SELECT col1 FROM [1,-2,3] ORDER BY abs(col1)",
        [{"col1": 1}, {"col1": -2}, {"col1": 3}],
    )
    eq_test_nrows(
        "SELECT col1 FROM [1,-2,3] ORDER BY abs(col1) DESC",
        [{"col1": 3}, {"col1": -2}, {"col1": 1}],
    )

    # order by (1 col, NULL)
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] ORDER BY 1",
        [{"col1": -2}, {"col1": 1}, {"col1": 3}, {"col1": NULL}],
    )
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] ORDER BY 1 NULLS LAST",
        [{"col1": -2}, {"col1": 1}, {"col1": 3}, {"col1": NULL}],
    )
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] ORDER BY 1 NULLS FIRST",
        [{"col1": NULL}, {"col1": -2}, {"col1": 1}, {"col1": 3}],
    )
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] ORDER BY 1 DESC",
        [{"col1": NULL}, {"col1": 3}, {"col1": 1}, {"col1": -2}],
    )
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] ORDER BY 1 DESC NULLS FIRST",
        [{"col1": NULL}, {"col1": 3}, {"col1": 1}, {"col1": -2}],
    )
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] ORDER BY 1 DESC NULLS LAST",
        [{"col1": 3}, {"col1": 1}, {"col1": -2}, {"col1": NULL}],
    )

    # order by (multi-cols)
    eq_test_nrows(
        "SELECT col1 as a, col2 as b FROM list(zip([1,2,3,1,2],[2,2,0,1,4])) ORDER BY 1"
        " ASC,  2 ASC",
        [
            {"a": 1, "b": 1},
            {"a": 1, "b": 2},
            {"a": 2, "b": 2},
            {"a": 2, "b": 4},
            {"a": 3, "b": 0},
        ],
    )
    eq_test_nrows(
        "SELECT col1 as a, col2 as b FROM list(zip([1,2,3,1,2],[2,2,0,1,4])) ORDER BY 1"
        " DESC, 2 DESC",
        list(
            reversed(
                [
                    {"a": 1, "b": 1},
                    {"a": 1, "b": 2},
                    {"a": 2, "b": 2},
                    {"a": 2, "b": 4},
                    {"a": 3, "b": 0},
                ]
            )
        ),
    )
    eq_test_nrows(
        "SELECT col1 as a, col2 as b FROM list(zip([1,2,3,1,2],[2,2,0,1,4])) ORDER BY 1"
        " ASC,  2 DESC",
        [
            {"a": 1, "b": 2},
            {"a": 1, "b": 1},
            {"a": 2, "b": 4},
            {"a": 2, "b": 2},
            {"a": 3, "b": 0},
        ],
    )
    eq_test_nrows(
        "SELECT col1 as a, col2 as b FROM list(zip([1,2,3,1,2],[2,2,0,1,4])) ORDER"
        " BY 2,1",
        [
            {"a": 3, "b": 0},
            {"a": 1, "b": 1},
            {"a": 1, "b": 2},
            {"a": 2, "b": 2},
            {"a": 2, "b": 4},
        ],
    )
    eq_test_nrows(
        "SELECT col1 as a, col2 as b FROM list(zip([1,2,3,1,2],[2,2,0,1,4])) ORDER BY 2"
        " DESC, 1 DESC",
        list(
            reversed(
                [
                    {"a": 3, "b": 0},
                    {"a": 1, "b": 1},
                    {"a": 1, "b": 2},
                    {"a": 2, "b": 2},
                    {"a": 2, "b": 4},
                ]
            )
        ),
    )

    # order by (with limit / offset / where)
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] WHERE col1 > 0 ORDER BY 1 DESC NULLS LAST",
        [{"col1": 3}, {"col1": 1}],
    )
    eq_test_nrows(
        "SELECT abs(col1) as col1 FROM [1,-2,NULL,3] ORDER BY 1 LIMIT 2",
        [{"col1": 1}, {"col1": 2}],
    )
    eq_test_nrows(
        "SELECT abs(col1) as col1 FROM [1,-2,NULL,3] ORDER BY 1 LIMIT 2 OFFSET 1",
        [{"col1": 2}, {"col1": 3}],
    )
    eq_test_nrows(
        "SELECT * FROM [1,-2,NULL,3] ORDER BY 1 LIMIT 0",
        [],
    )


def test_agg():
    # aggregate functions (overall)
    funcs = (
        # sql func, python func, remove nulls on python func?
        ("sum_agg(col1)", lambda x: sum(x) if x else NULL, True),
        (
            "prod_agg(col1)",
            lambda x: reduce(lambda a, b: a * b, x) if x else NULL,
            True,
        ),
        ("count_agg(col1)", len, True),
        ("count_agg(*)", len, False),
        ("avg_agg(col1)", lambda x: sum(x) / len(x) if x else NULL, True),
        ("min_agg(col1)", lambda x: min(x) if x else NULL, True),
        ("max_agg(col1)", lambda x: max(x) if x else NULL, True),
        ("list_agg(col1)", lambda x: list(x), False),
        ("list_agg(col1, False)", lambda x: list(x), True),
        ('string_agg(col1,",")', lambda x: ",".join(map(str, x)), True),
        ('string_agg(col1,",",True)', lambda x: ",".join(map(str, x)), False),
        (
            "sorted(list(set_agg(col1)), key=lambda x: (x is NULL, x))",
            lambda y: sorted(list(set(y)), key=lambda x: (x is NULL, x)),
            False,
        ),
        ("sorted(list(set_agg(col1, False)))", lambda y: sorted(list(set(y))), True),
        (
            "dict_agg(str(col1), row_number)",
            lambda y: {str(k): v + 1 for v, k in enumerate(y) if k is not NULL},
            False,
        ),
        ("first_agg(col1)", lambda x: x[0] if x else NULL, False),
        ("first_agg(col1, False)", lambda x: x[0] if x else NULL, True),
        ("last_agg(col1)", lambda x: x[-1] if x else NULL, False),
        ("last_agg(col1, False)", lambda x: x[-1] if x else NULL, True),
        ("lag_agg(col1)", lambda x: x[-2] if len(x) > 1 else NULL, False),
        ("lag_agg(col1,2)", lambda x: x[-3] if len(x) > 2 else NULL, False),
        ("count_distinct_agg(col1)", lambda x: len(set(x)), True),
        ("count_distinct_agg(*)", lambda x: len(set(x)), False),
        (
            "any_agg(col1 == 1)",
            lambda x: any(map(lambda y: y == 1, x)) if x else NULL,
            True,
        ),
        (
            "every_agg(col1 > 0)",
            lambda x: all(map(lambda y: y > 0, x)) if x else NULL,
            True,
        ),
    )

    tst_lists = [
        [NULL],
        [NULL, NULL, NULL, NULL],
        [NULL, 11, NULL],
        [NULL, 11, 5, 10, NULL, 3, 3, 10, 4],
        [12],
        range(1, 21),
        range(-10, 6),
        [int(math.cos(x) * 100) / 100.0 for x in range(100)],
    ]
    for tst_list in tst_lists:
        tst_list_clean = list(filter(lambda x: x is not NULL, tst_list))
        for sql_func, tst_func, ignore_nulls in funcs:
            col_name = sql_func[:5]
            lst = tst_list_clean if ignore_nulls else tst_list
            eq_test_1row(
                f"SELECT {sql_func} as {col_name} FROM {tst_list}",
                {col_name: tst_func(lst)},
            )
            if not sql_func.startswith(
                "dict_agg"
            ):  # dicts do not suport the * operator
                eq_test_1row(
                    f"SELECT {sql_func} as a, {sql_func}*2 as b FROM {tst_list}",
                    {"a": tst_func(lst), "b": tst_func(lst) * 2},
                )

    # this would return a row with 0 in standard SQL, but in SpyQL returns no rows
    eq_test_nrows("SELECT count(*) FROM []", [])

    # partials
    for tst_list in tst_lists:
        eq_test_nrows(
            "SELECT PARTIALS list_agg(col1) as a, count_agg(col1) as c1, count_agg(*)"
            f" as c2, first_agg(col1) as f, lag_agg(col1) as l FROM {tst_list}",
            [
                {
                    "a": list(tst_list[:n]),
                    "c1": len(list(filter(lambda el: el is not NULL, tst_list[:n]))),
                    "c2": n,
                    "f": tst_list[0],
                    "l": NULL if n < 2 else tst_list[n - 2],
                }
                for n in range(1, len(tst_list) + 1)
            ],
        )


def test_groupby():
    eq_test_1row("SELECT 1 as a FROM range(1) GROUP BY col1", {"a": 1})
    eq_test_1row(
        "SELECT 1 as a, count_agg(*) as c FROM range(1) GROUP BY 1", {"a": 1, "c": 1}
    )
    eq_test_1row(
        "SELECT 1 as a, count_agg(*) as c FROM range(10) GROUP BY 1", {"a": 1, "c": 10}
    )
    eq_test_1row(
        "SELECT 1 as a, 2 as b, count_agg(*) as c FROM range(100) GROUP BY 1, 2",
        {"a": 1, "b": 2, "c": 100},
    )
    eq_test_nrows(
        "SELECT col1 % 2 as a, 2 as b, count_agg(*) as c, min_agg(col1) as mn,"
        " max_agg(col1) as mx FROM range(101) GROUP BY 1, 1+1",
        [
            {"a": 0, "b": 2, "c": 51, "mn": 0, "mx": 100},
            {"a": 1, "b": 2, "c": 50, "mn": 1, "mx": 99},
        ],
    )
    eq_test_nrows(
        "SELECT col1 % 3 as a, 2 as b, max_agg(col1) as mx FROM range(100) GROUP BY 1,"
        " 2 ORDER BY 1 DESC",
        [
            {"a": 2, "b": 2, "mx": 98},
            {"a": 1, "b": 2, "mx": 97},
            {"a": 0, "b": 2, "mx": 99},
        ],
    )
    eq_test_nrows(
        "SELECT col1 % 3 as a, 2 as b, max_agg(col1) as mx FROM range(100) GROUP BY 1,"
        " 2 ORDER BY max_agg(col1)",
        [
            {"a": 1, "b": 2, "mx": 97},
            {"a": 2, "b": 2, "mx": 98},
            {"a": 0, "b": 2, "mx": 99},
        ],
    )


def test_distinct():
    eq_test_1row("SELECT DISTINCT 1 as a FROM range(1)", {"a": 1})
    eq_test_1row("SELECT DISTINCT 1 as a FROM range(10)", {"a": 1})
    eq_test_1row("SELECT DISTINCT 1 as a, 2 as b FROM range(100)", {"a": 1, "b": 2})
    eq_test_nrows(
        "SELECT DISTINCT col1 % 2 as a, 2 as b FROM range(100)",
        [{"a": 0, "b": 2}, {"a": 1, "b": 2}],
    )
    eq_test_nrows(
        "SELECT DISTINCT col1 % 3 as a, 2 as b FROM range(100) ORDER BY 1 DESC",
        [{"a": 2, "b": 2}, {"a": 1, "b": 2}, {"a": 0, "b": 2}],
    )
    eq_test_nrows(
        "SELECT DISTINCT -(col1%3) as a, -(col1%2) as b FROM range(90) ORDER BY 1,2",
        [
            {"a": -2, "b": -1},
            {"a": -2, "b": 0},
            {"a": -1, "b": -1},
            {"a": -1, "b": 0},
            {"a": 0, "b": -1},
            {"a": 0, "b": 0},
        ],
    )

    # Distinct jsons
    res = run_cli(
        "SELECT DISTINCT json FROM json EXPLODE json.a TO json",
        data=(
            '{"a": [1, 2, 2], "b": "three"}\n{"a": [], "b": "none"}\n{"a": [4], "b":'
            ' "four"}\n'
        ),
    )
    assert json_output(res.output) == [
        {"a": 1, "b": "three"},
        {"a": 2, "b": "three"},
        {"a": 4, "b": "four"},
    ]
    assert res.exit_code == 0


def test_null():
    eq_test_1row("SELECT NULL", {"NULL": NULL})
    eq_test_1row("SELECT NULL+1", {"NULL_1": NULL})
    eq_test_1row("SELECT int('')", {"int": NULL})
    eq_test_1row("SELECT coalesce(NULL,2)", {"coalesce_NULL_2": 2})
    eq_test_1row("SELECT coalesce(3,2)", {"coalesce_3_2": 3})
    eq_test_1row("SELECT nullif(1,1)", {"nullif_1_1": NULL})
    eq_test_1row("SELECT nullif(3,2)", {"nullif_3_2": 3})


def test_processors():
    # JSON input and NULLs
    for jsonproc in ["json", "orjson"]:
        eq_test_1row(
            f"SELECT json.a FROM {jsonproc}",
            {"a": 1},
            data='{"a": 1}\n',
        )
        eq_test_1row(
            f"SELECT json.a FROM {jsonproc}",
            {"a": NULL},
            data='{"a": null}\n',
        )
        eq_test_1row(
            f"SELECT json.b FROM {jsonproc}",
            {"b": NULL},
            data='{"a": 1}\n',
        )

        # JSON EXPLODE
        eq_test_nrows(
            f"SELECT json.a, json.b FROM {jsonproc} EXPLODE json.a",
            [
                {"a": 1, "b": "three"},
                {"a": 2, "b": "three"},
                {"a": 3, "b": "three"},
                {"a": 4, "b": "four"},
            ],
            data=(
                '{"a": [1, 2, 3], "b": "three"}\n{"a": [], "b": "none"}\n{"a": [4], "b":'
                ' "four"}\n'
            ),
        )
        eq_test_nrows(f"SELECT * FROM {jsonproc}", [], data="")

    # CSV input and NULLs
    eq_test_nrows(
        "SELECT a as a FROM csv",
        [{"a": 1}, {"a": 4}, {"a": 7}],
        data="a,b,c\n1,2,3\n4,5,6\n7,8,9",
    )
    eq_test_nrows(
        "SELECT a as a FROM csv",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data="a,b,c\n,2,3\n4,5,6\n,8,9",
    )
    eq_test_nrows(
        "SELECT a as a FROM csv",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data="a,b,c\n,2,3\n4,5,6\n,8,9",
        input_options={"delimiter": ","},
    )
    eq_test_nrows(
        "SELECT int(a) as a FROM csv",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data="a,b,c\n,2,3\n4,5,6\noops,8,9",
        input_options={"delimiter": ",", "infer_dtypes": False},
    )
    eq_test_nrows(
        "SELECT a as a FROM csv",
        [{"a": ""}, {"a": "4"}, {"a": ""}],
        data="a,b,c\n,2,3\n4,5,6\n,8,9",
        input_options={"delimiter": ",", "infer_dtypes": False},
    )
    eq_test_nrows(
        "SELECT col1 as a FROM csv",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data=",2,3\n4,5,6\n,8,9",
        input_options={"delimiter": ",", "header": False},
    )
    eq_test_nrows(
        "SELECT col1 as a, col2 as b, col3 as c FROM csv",  # type inference test
        [
            {"a": NULL, "b": 2.0, "c": "3"},
            {"a": 4, "b": 5.0, "c": "ola"},
            {"a": NULL, "b": NULL, "c": ""},
        ],
        data=",2,3\n4,5.0,ola\n,,",
        input_options={"delimiter": ",", "header": False},
    )
    eq_test_nrows("SELECT * FROM csv", [], data="")

    # Text input and NULLs
    eq_test_nrows(
        "SELECT int(col1) as a FROM text",
        [{"a": 1}, {"a": 4}, {"a": 7}],
        data="1\n4\n7",
    )
    eq_test_nrows(
        "SELECT int(col1) as a FROM text",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data="\n4\noops",
    )
    eq_test_nrows("SELECT * FROM text", [], data="")

    # SPy input and NULLs
    eq_test_nrows(
        "SELECT a as a FROM spy",
        [{"a": 1}, {"a": 4}, {"a": 7}],
        data="".join(
            [
                SpyWriter.pack(line)
                for line in [["a", "b", "c"], [1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ]
        ),
    )
    eq_test_nrows(
        "SELECT int(a) as a FROM spy",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data="".join(
            [
                SpyWriter.pack(line)
                for line in [["a", "b", "c"], [NULL, 2, 3], [4, 5, 6], ["oops", 8, 9]]
            ]
        ),
    )
    eq_test_nrows(
        "SELECT a as a FROM spy",
        [{"a": {"aa": [11, 12, 13]}}, {"a": 4}, {"a": "ok"}],
        data="".join(
            [
                SpyWriter.pack(line)
                for line in [
                    ["a", "b", "c"],
                    [{"aa": [11, 12, 13]}, 2, 3],
                    [4, 5, 6],
                    ["ok", 8, 9],
                ]
            ]
        ),
    )
    eq_test_nrows("SELECT * FROM spy", [], data="")


def test_row_access():
    # JSON input and NULLs
    eq_test_1row("SELECT row.a FROM json", {"a": 1}, data='{"a": 1}\n')
    eq_test_1row("SELECT row.a FROM json", {"a": NULL}, data='{"a": null}\n')
    eq_test_1row("SELECT row.b FROM json", {"b": NULL}, data='{"a": 1}\n')

    # JSON EXPLODE
    eq_test_nrows(
        "SELECT row.a, row.b FROM json EXPLODE row.a",
        [
            {"a": 1, "b": "three"},
            {"a": 2, "b": "three"},
            {"a": 3, "b": "three"},
            {"a": 4, "b": "four"},
        ],
        data=(
            '{"a": [1, 2, 3], "b": "three"}\n{"a": [], "b": "none"}\n{"a": [4], "b":'
            ' "four"}\n'
        ),
    )

    # CSV input
    eq_test_nrows(
        "SELECT row.a FROM csv",
        [{"a": 1}, {"a": 4}, {"a": 7}],
        data="a,b,c\n1,2,3\n4,5,6\n7,8,9",
    )
    eq_test_nrows(
        "SELECT row.a FROM csv",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data="a,b,c\n,2,3\n4,5,6\n,8,9",
    )

    # Text input and NULLs
    eq_test_nrows(
        "SELECT int(row.col1) as a FROM text",
        [{"a": 1}, {"a": 4}, {"a": 7}],
        data="1\n4\n7",
    )

    # SPy input and NULLs
    eq_test_nrows(
        "SELECT row.a FROM spy",
        [{"a": 1}, {"a": 4}, {"a": 7}],
        data="".join(
            [
                SpyWriter.pack(line)
                for line in [["a", "b", "c"], [1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ]
        ),
    )
    eq_test_nrows(
        "SELECT int(row.a) as a FROM spy",
        [{"a": NULL}, {"a": 4}, {"a": NULL}],
        data="".join(
            [
                SpyWriter.pack(line)
                for line in [["a", "b", "c"], [NULL, 2, 3], [4, 5, 6], ["oops", 8, 9]]
            ]
        ),
    )

    # Python processor
    eq_test_nrows(
        'SELECT row.a FROM [{"a": 1, "b": 3}, {"a": 4}, {"a": 7}]',
        [{"a": 1}, {"a": 4}, {"a": 7}],
    )


def test_metadata():
    eq_test_nrows(
        "SELECT cols, _values, _names, row FROM csv",
        [
            {
                "cols": [NULL, 2, 3],
                "_values": ["", "2", "3"],
                "_names": ["a", "b", "c"],
                "row": {"a": NULL, "b": 2, "c": 3},
            },
            {
                "cols": [4, 5, 6],
                "_values": ["4", "5", "6"],
                "_names": ["a", "b", "c"],
                "row": {"a": 4, "b": 5, "c": 6},
            },
            {
                "cols": [NULL, 8, 9],
                "_values": ["", "8", "9"],
                "_names": ["a", "b", "c"],
                "row": {"a": NULL, "b": 8, "c": 9},
            },
        ],
        data="a,b,c\n,2,3\n4,5,6\n,8,9",
    )
    eq_test_nrows(
        "SELECT cols, _values, _names, row FROM csv",
        [
            {
                "cols": [NULL, 2, 3],
                "_values": ["", "2", "3"],
                "_names": ["col1", "col2", "col3"],
                "row": {"col1": NULL, "col2": 2, "col3": 3},
            },
            {
                "cols": [4, 5, 6],
                "_values": ["4", "5", "6"],
                "_names": ["col1", "col2", "col3"],
                "row": {"col1": 4, "col2": 5, "col3": 6},
            },
            {
                "cols": [NULL, 8, 9],
                "_values": ["", "8", "9"],
                "_names": ["col1", "col2", "col3"],
                "row": {"col1": NULL, "col2": 8, "col3": 9},
            },
        ],
        data=",2,3\n4,5,6\n,8,9",
        input_options={"delimiter": ",", "header": False},
    )
    eq_test_1row(
        "SELECT cols, _values, _names, row FROM json",
        {
            "cols": [{"a": 1}],
            "_values": [{"a": 1}],
            "_names": ["json"],
            "row": {"a": 1},
        },
        data='{"a": 1}\n',
    )
    eq_test_1row(
        "SELECT cols, _values, _names, row FROM text",
        {
            "cols": ["hello"],
            "_values": ["hello"],
            "_names": ["col1"],
            "row": {"col1": "hello"},
        },
        data="hello\n",
    )
    eq_test_nrows(
        "SELECT cols, _values, _names, row FROM spy",
        [
            {
                "cols": [1, NULL, 3],
                "_values": [1, NULL, 3],
                "_names": ["a", "b", "c"],
                "row": {"a": 1, "b": NULL, "c": 3},
            },
            {
                "cols": [4, 5, 6],
                "_values": [4, 5, 6],
                "_names": ["a", "b", "c"],
                "row": {"a": 4, "b": 5, "c": 6},
            },
            {
                "cols": [7, 8, 9],
                "_values": [7, 8, 9],
                "_names": ["a", "b", "c"],
                "row": {"a": 7, "b": 8, "c": 9},
            },
        ],
        data="".join(
            [
                SpyWriter.pack(line)
                for line in [["a", "b", "c"], [1, NULL, 3], [4, 5, 6], [7, 8, 9]]
            ]
        ),
    )


def test_custom_syntax():
    # easy access to dic fields using ->
    eq_test_1row(
        "SELECT col1->three * 2 as six, col1->'twenty one' + 3 AS twentyfour,"
        " col1->hello->world.upper() AS caps "
        "FROM [[{'three': 3, 'twenty one': 21,"
        " 'hello':{'world': 'hello world'}}]] "
        "WHERE col1->three > 0 "
        "ORDER BY col1->three",
        {"six": 6, "twentyfour": 24, "caps": "HELLO WORLD"},
    )

    # easy access to dic fields using .
    eq_test_1row(
        "SELECT col1.three * 2 as six, col1.hello.world.upper() AS caps "
        "FROM {'three': 3, 'twenty one': 21,"
        " 'hello': {'world': 'hello world'}} "
        "WHERE col1.three > 0 "
        "ORDER BY col1.three",
        {"six": 6, "caps": "HELLO WORLD"},
    )

    eq_test_1row(
        """
        SELECT
            .one,
            .two.a,
            json.two.b,
            ._u_,
            .1 as dot1,
            0.2 as dot2,
            [2,1,2,1,2].count(2) as rsb,
            (3,3).count(3) as rrb,
            len({}.keys()) as rcb,
            (.one + 2) * 2 as lrb,
            (10,20,30)[.one] as lsb,
            {.one:'ok', 0:'bad'}[.one] as lcb,
            sum([  .one, .one,.one]) as three,
            {'a':.one }['a']+.one*.one/.one**.one-.one/.1 as arith,
            ._u_.count(200)==.one as und,
            'a'.upper() as str1,
            "b".upper() as str2,
            list(.keys()) as keys
        FROM json
        """,
        {
            "one": 1,
            "two_a": 12,
            "two_b": 13,
            "_u_": [100, 200, 300],
            "dot1": 0.1,
            "dot2": 0.2,
            "rsb": 3,
            "rrb": 2,
            "rcb": 0,
            "lrb": 6,
            "lsb": 20,
            "lcb": "ok",
            "three": 3,
            "arith": -8.0,
            "und": True,
            "str1": "A",
            "str2": "B",
            "keys": ["one", "two", "_u_"],
        },
        data='{"one": 1, "two": {"a": 12, "b": 13}, "_u_": [100,200,300] }\n',
    )

    eq_test_1row(
        """
        SELECT
            len(.),
            {'all':.},
            .,
            list((.,)),
            [.,.],
            'one' in .
        FROM json
        """,
        {
            "len_row": 2,
            "all_row": {"all": {"one": 1, "two": 2}},
            "row": {"one": 1, "two": 2},
            "list_row": [{"one": 1, "two": 2}],
            "row_row": [{"one": 1, "two": 2}, {"one": 1, "two": 2}],
            "one_in_row": True,
        },
        data='{"one": 1, "two": 2}\n',
    )


def test_errors():
    # TODO find way to test custom error output
    exception_test("SELECT 2 + ''", TypeError)
    exception_test("SELECT abc", NameError)
    exception_test("SELECT ,1", SyntaxError)
    exception_test("SELECT 'abcde ", SyntaxError)
    exception_test("SELECT 1 SELECT 2", SyntaxError)
    exception_test("SELECT 1 WHERE True FROM [1]", SyntaxError)
    exception_test("WHERE True", SyntaxError)
    exception_test("SELECT 1 TO _this_writer_does_not_exist_", SyntaxError)
    exception_test("SELECT 1 FROM [1,2,,]]", SyntaxError)
    exception_test("IMPORT _this_module_does_not_exist_ SELECT 1", ModuleNotFoundError)
    exception_test(
        "SELECT 1 TO csv", TypeError, output_options={"nonexisting_option": 1}
    )
    exception_test(
        "SELECT 1 TO plot", TypeError, output_options={"nonexisting_option": 1}
    )
    exception_test(
        "SELECT 1 FROM csv", TypeError, input_options={"nonexisting_option": 1}
    )
    exception_test(
        "SELECT 1 FROM spy", TypeError, input_options={"nonexisting_option": 1}
    )
    exception_test("SELECT 1 TO csv", TypeError, output_options={"delimiter": "bad"})
    exception_test("SELECT 1 FROM csv", TypeError, input_options={"delimiter": "bad"})
    exception_test("SELECT 1 TO csv", SystemExit, output_options={"la": None})
    exception_test("SELECT int('abcde')", ValueError, warning_flag="error")
    exception_test("SELECT float('')", ValueError, warning_flag="error")
    exception_test("SELECT float('')", ValueError, warning_flag="error")
    exception_test("SELECT DISTINCT count_agg(1)", SyntaxError)
    exception_test("SELECT count_agg(1) GROUP BY 1", SyntaxError)
    exception_test("SELECT 1 FROM range(3) WHERE max_agg(col1) > 0", SyntaxError)
    exception_test(
        "SELECT row.a FROM [{'a':1},{'a':2},{'a':3}] EXPLODE row.a", TypeError
    )


def test_sql_output():
    """
    Writes to an in memory sqlite DB and reads back to test
    """
    conn = sqlite3.connect(":memory:")
    conn.cursor().execute(
        """CREATE TABLE test1(
        aint int,
        afloat numeric(2,1),
        aintnull int,
        astrnull text,
        astr text,
        alist text,
        adict text)
    """
    )

    query = """
        SELECT
            col1 as aint,
            col1/2 + 0.01 as afloat,
            NULL if col1==2 else 100 as aintnull,
            NULL if col1==1 else 'hello' as astrnull,
            'abc' + str(col1) as astr,
            [1,2,3] as alist,
            {'a':col1, 'a2': col1*2} as adict
        FROM [1,2,3]
        TO sql
        """

    res = run_cli(query, ["-Otable=test1", "-Ochunk_size=2"])
    assert res.exit_code == 0
    for insert_stat in res.output.splitlines():
        # run inserts from spyql in sqlite
        conn.cursor().execute(insert_stat)

    result = conn.cursor().execute("select * from test1").fetchall()
    conn.close()
    expectation = [
        (1, 0.51, 100, None, "abc1", "[1, 2, 3]", "{'a': 1, 'a2': 2}"),
        (2, 1.01, None, "hello", "abc2", "[1, 2, 3]", "{'a': 2, 'a2': 4}"),
        (3, 1.51, 100, "hello", "abc3", "[1, 2, 3]", "{'a': 3, 'a2': 6}"),
    ]
    assert expectation == result


def test_plot_output():
    # just checking that it does not break...
    res = run_cli("SELECT col1 as abc, col1*2 FROM range(20) TO plot")
    assert res.exit_code == 0
    res = run_cli("SELECT col1 FROM [1,2,NULL,3,None,4] TO plot")
    assert res.exit_code == 0



import csv
import pickle
import io
import sys
import re
import os
from itertools import islice, chain
from io import StringIO
import copy

from spyql.writer import Writer
from spyql.output_handler import OutputHandler
from spyql.query_result import QueryResult
import spyql.nulltype
import spyql.sqlfuncs
from spyql.qdict import qdict
import spyql.log
from spyql.utils import make_str_valid_varname, isiterable, is_row_collapsable
import spyql.agg


def init_vars(user_query_vars={}):
    """Initializes dict of variables for user queries"""
    vars = dict()
    # imports for user queries (TODO move to init.py when mature)
    exec(
        "from datetime import datetime, date, timezone\n"
        "from spyql.nulltype import *\n"
        "from spyql.sqlfuncs import *\n"
        "from spyql.qdict import *\n"
        "from spyql.agg import *\n"
        "from math import *\n"
        "import re\n",
        {},
        vars,
    )

    try:
        # user defined imports, functions, etc
        config_home = os.environ.get(
            "XDG_CONFIG_HOME", os.path.expanduser(os.path.join("~", ".config"))
        )
        init_fname = os.path.join(config_home, "spyql", "init.py")
        with open(init_fname) as f:
            exec(f.read(), {}, vars)
            spyql.log.user_debug(f"Succesfully loaded {init_fname}")
    except FileNotFoundError:
        spyql.log.user_debug(f"Init file not found: {init_fname}")
    except Exception as e:
        spyql.log.user_warning(f"Could not load {init_fname}", e)

    # update the accessible vars with user defined vars, if overlap, warn the user
    for x in set(vars.keys()) & set(user_query_vars.keys()):
        spyql.log.user_warning(
            f"Overloading builtin name '{x}', somethings may not work!"
        )
    vars.update(user_query_vars)

    return vars


class Processor:
    @staticmethod
    def input_processors():
        return {
            "JSON": JSONProcessor,
            "ORJSON": ORJSONProcessor,
            "CSV": CSVProcessor,
            "TEXT": TextProcessor,
            "SPY": SpyProcessor,
        }

    @staticmethod
    def make_processor(prs, strings, input_options={}):
        """
        Factory for making an input processor based on the parsed query
        """
        try:
            from_clause = prs["from"]
            processor_name = ""
            if not from_clause:  # no from close, single select eval
                processor_name = "default"
                return Processor(prs, strings, **input_options)
            elif isinstance(from_clause, dict):  # there's an input data processor
                processor_name = from_clause["name"]
                processor = Processor.input_processors()[processor_name.upper()]
                input_options.update(from_clause["kwargs"])
                return processor(prs, strings, *from_clause["args"], **input_options)
            else:  # python expression
                processor_name = "python"
                return PythonExprProcessor(prs, strings, **input_options)
        except TypeError as e:
            spyql.log.user_error(f"Could not create '{processor_name}' processor", e)

    def __init__(self, prs, strings, path=None):
        spyql.log.user_debug(f"Loading {self.__class__.__name__}")
        self.prs = prs  # parsed query
        spyql.log.user_debug(self.prs)
        self.strings = strings  # quoted strings
        self.path = path
        try:
            self.input_file = open(path, "r") if path else sys.stdin
        except FileNotFoundError as e:
            spyql.log.user_error(f"Input file not found: {path}", e)
        except Exception as e:
            spyql.log.user_error(f"Could not load {path}", e)

        self.input_col_names = []  # column names of the input data
        self.translations = copy.deepcopy(
            spyql.sqlfuncs.NULL_SAFE_FUNCS
        )  # map for alias, functions to be renamed...
        self.has_header = False
        self.casts = dict()
        self.col_values_exprs = []
        self.writer = None
        self.query_has_reference2row = prs["hints"]["has_reference2row"]

    def close(self):
        if self.path:
            self.input_file.close()

    def reading_data(self):
        """
        Returns True after reading header, metadata, etc in input file
        """
        return True

    def handle_header_row(self, row):
        """
        Action for header row (e.g. column name definition)
        """
        pass

    def handle_1st_data_row(self, row):
        """
        Action for handling the first row of data
        """
        self.n_input_cols = len(row) if row else 0

        default_col_names = [
            self.default_col_name(_i) for _i in range(self.n_input_cols)
        ]
        self.col_values_exprs = [
            f"{self.casts[_i]}(_values[{_i}])" if _i in self.casts else f"_values[{_i}]"
            for _i in range(self.n_input_cols)
        ]
        # dictionary to translate col names to accesses to `_values`
        self.translations.update(dict(zip(default_col_names, self.col_values_exprs)))
        if self.input_col_names:
            # TODO check if len(input_col_names) == self.n_input_cols
            self.translations.update(
                dict(zip(self.input_col_names, self.col_values_exprs))
            )

        # metadata: list of column names
        _names = self.input_col_names if self.input_col_names else default_col_names
        self.vars["_names"] = _names
        # list of [col1,col2,...]
        cols_expr = "[" + ",".join(self.col_values_exprs) + "]"
        self.translations["cols"] = cols_expr

        # code for instantiating the `row` variable, a dict of {col1: value1, ...} }
        # if the result is a single column of type dict, then returns that dict instead
        # TODO extend collapsing to Pandas, NumPy arrays, etc
        row_expr = (
            self.col_values_exprs[0]
            if is_row_collapsable(row, _names)
            else f"qdict(zip(_names, {cols_expr}))"
        )

        self.row_expr = compile(row_expr, "row_expr", "eval")

    def make_row_meta(self):
        return eval(self.row_expr, self.vars, self.vars)

    def make_out_cols_names(self, out_cols_names):
        """
        Creates list of output column names
        """
        input_col_names = self.input_col_names
        if not input_col_names:
            input_col_names = [
                self.default_col_name(i) for i in range(self.n_input_cols)
            ]
        out_cols_names = [
            (input_col_names if name == "*" else [name]) for name in out_cols_names
        ]
        out_cols_names = [
            name for sublist in out_cols_names for name in sublist
        ]  # flatten
        return out_cols_names

    def get_input_iterator(self):
        """
        Returns iterator over input (e.g. list if rows)
        Each row is list with one value per column
        e.g.::

            [[1] ,[2], [3]]               # 3 rows with a single col
            [[1,'a'], [2,'b'], [3,'c']]   # 3 rows with 2 cols
        """
        return [[None]]  # default: returns a single line with a 'null' column

    def default_col_name(self, idx):
        """
        Default column names, e.g. col1 for the first column
        """
        return f"col{idx+1}"

    def prepare_expression(self, expr):
        """
        Replaces identifiers (column names) in sql expressions by references to
        `_values` and put (quoted) strings back
        """
        if expr == "*":
            return self.col_values_exprs

        if isinstance(expr, int):
            # special case: expression is out col number (1-based)
            return [f"_res[{expr-1}]"]  # reuses existing result

        for id, replacement in self.translations.items():
            pattern = rf"(?<![\w\.])({id})\b"
            expr = re.compile(pattern).sub(replacement, expr)

        return [self.strings.put_strings_back(expr)]

    def is_clause_single(self, clause):
        """
        True if clause can only have a single expression
        """
        return clause not in {
            "select",
            "group by",
            "order by",
        }

    def compile_clause(self, clause, clause_modifier=None, mode="eval"):
        """
        Compiles a clause of the query
        """
        prs_clause = self.prs[clause]
        if not prs_clause:
            return None  # empty clause

        if clause_modifier:
            prs_clause = clause_modifier.format(prs_clause)

        single = self.is_clause_single(clause)
        clause_exprs = None
        if single:  # a clause with a single expression like WHERE
            clause_exprs = self.prepare_expression(prs_clause)
            if len(clause_exprs) > 1:
                spyql.log.user_error(
                    f"could not compile {clause.upper()} clause",
                    SyntaxError(
                        f"{clause.upper()} clause should not have more than 1"
                        " expression"
                    ),
                )
            clause_exprs = clause_exprs[0]
        else:  # a clause with multiple expressions like SELECT
            clause_exprs = [self.prepare_expression(c["expr"]) for c in prs_clause]
            clause_exprs = [
                item for sublist in clause_exprs for item in sublist
            ]  # flatten (because of '*')
            clause_exprs = ",".join(clause_exprs) + ","  # tuple constructor

        try:
            return compile(clause_exprs, f"<{clause}>", mode)
        except Exception as main_exception:
            if not single:
                # breaks down clause into expressions and tries
                # compiling one by one to detect in which expression
                # the error happened
                for c in range(len(prs_clause)):
                    try:
                        expr = prs_clause[c]["expr"]
                        translation = self.prepare_expression(expr)
                        for trans in translation:
                            if not trans.strip():
                                raise SyntaxError("empty expression")
                            compile(trans, f"<{clause}>", mode)
                    except Exception as expr_exception:
                        spyql.log.user_error(
                            f"could not compile {clause.upper()} expression #{c+1}",
                            expr_exception,
                            self.strings.put_strings_back(expr),
                        )

            spyql.log.user_error(
                f"could not compile {clause.upper()} clause", main_exception
            )

    def eval_clause(self, clause, clause_exprs, mode="eval"):
        """
        Evaluates/executes a previously compiled clause
        """
        if not clause_exprs:
            return
        cmd = eval if mode == "eval" else exec
        try:
            return cmd(clause_exprs, self.vars, self.vars)

        except Exception as main_exception:
            # this code is useful for debugging and not the actual processing
            prs_clause = self.prs[clause]
            if not self.is_clause_single(clause):
                # breaks down clause into expressions and tries
                # evaluating/executing one by one to detect
                # in which expression the error happened
                for c in range(len(prs_clause)):
                    try:
                        expr = prs_clause[c]["expr"]
                        spyql.log.user_debug("expression", expr)
                        translation = self.prepare_expression(expr)
                        for trans in translation:
                            spyql.log.user_debug("translated expression", trans)
                            cmd(trans, self.vars, self.vars)
                    except Exception as expr_exception:
                        spyql.log.user_error(
                            f"could not evaluate {clause.upper()} expression #{c+1}",
                            expr_exception,
                            self.strings.put_strings_back(expr),
                            self.vars,
                        )

            spyql.log.user_error(
                f"could not evaluate {clause.upper()} clause",
                main_exception,
                vars=self.vars,
            )

    # main
    def go(self, output_options, user_query_vars={}) -> QueryResult:
        output_handler = OutputHandler.make_handler(self.prs)
        self.writer = Writer.make_writer(self.prs["to"], output_options)
        output_handler.set_writer(self.writer)
        nrows_in = self._go(output_handler, user_query_vars)
        output_handler.finish()
        spyql.log.user_info("#rows  in", nrows_in)
        spyql.log.user_info("#rows out", output_handler.rows_written)

        stats = {"rows_in": nrows_in, "rows_out": output_handler.rows_written}
        return self.writer.result(), stats

    def _go(self, output_handler, user_query_vars):
        select_expr = None
        where_expr = None
        explode_expr = None
        explode_cmd = None
        explode_its = [None]  # 1 element by default (no explosion)
        groupby_expr = None
        orderby_expr = None
        _values = []
        _res = tuple()
        _group_res = tuple()
        _sort_res = tuple()
        row_number = 0
        input_row_number = 0

        self.vars = init_vars(user_query_vars)
        spyql.agg._init_aggs()

        # import user modules
        self.eval_clause(
            "import",
            self.compile_clause("import", "import {}", mode="exec"),
            mode="exec",
        )

        # gets user-defined output cols names (with AS alias)
        out_cols_names = [c["name"] for c in self.prs["select"]]

        # should not accept more than 1 source, joins, etc (at least for now)
        for _values in self.get_input_iterator():
            input_row_number = input_row_number + 1

            self.vars["_values"] = _values
            self.vars["input_row_number"] = input_row_number

            if not self.reading_data():
                self.handle_header_row(_values)
                continue

            # print header
            if select_expr is None:  # 1st input data row
                self.handle_1st_data_row(_values)
                output_handler.writer.writeheader(
                    self.make_out_cols_names(out_cols_names)
                )
                if output_handler.is_done():
                    return 0  # in case of `limit 0`

                select_expr = self.compile_clause("select")
                where_expr = self.compile_clause("where")
                explode_expr = self.compile_clause("explode")
                groupby_expr = self.compile_clause("group by")
                orderby_expr = self.compile_clause("order by")

            if self.query_has_reference2row:
                # only builds the row variable if there is a reference to it
                self.vars["row"] = self.make_row_meta()

            if explode_expr:
                explode_its = self.eval_clause("explode", explode_expr)
                if not isiterable(explode_its):
                    spyql.log.user_error(
                        "Invalid EXPLODE clause",
                        TypeError(
                            f"{self.prs['explode']} has type {type(explode_its)}, which"
                            " is not iterable"
                        ),
                        vars=self.vars,
                    )
            for explode_it in explode_its:
                if explode_expr:
                    self.vars["explode_it"] = explode_it
                    if not explode_cmd:
                        explode_cmd = self.compile_clause(
                            "explode", "{} = explode_it", mode="exec"
                        )
                    self.eval_clause("explode", explode_cmd, mode="exec")
                    self.vars["_values"] = _values

                # filter (opt: eventually could be done before exploding)
                if not where_expr or self.eval_clause("where", where_expr):
                    # input line is eligible
                    row_number = row_number + 1
                    self.vars["row_number"] = row_number

                    if groupby_expr:
                        # group by can ref output columns, but does not depend on the
                        # execution of the select clause: refs to output columns are
                        # replaced by the correspondent expression
                        _group_res = self.eval_clause("group by", groupby_expr)
                        # we need to set the group key before running the select because
                        # aggregate functions need to know the group key beforehand
                        spyql.agg._start_new_agg_row(_group_res)

                    # calculate outputs
                    _res = self.eval_clause("select", select_expr)

                    if orderby_expr:
                        # in the order by clause, references to output columns use the
                        # outputs of the evaluation of the select expression
                        self.vars["_res"] = _res
                        _sort_res = self.eval_clause("order by", orderby_expr)

                    is_done = output_handler.handle_result(
                        _res, _sort_res, _group_res
                    )  # deal with output
                    if is_done:
                        # e.g. when reached limit
                        return input_row_number - (1 if self.has_header else 0)

        return input_row_number - (1 if self.has_header else 0)


class PythonExprProcessor(Processor):
    def __init__(self, prs, strings):
        super().__init__(prs, strings)

    # input is a Python expression or a ref that is passed in the vars.
    def get_input_iterator(self):
        spyql.log.user_debug(f"Trying to read as python expression")
        e = self.eval_clause("from", self.compile_clause("from"))
        if e:
            if not isiterable(e):
                e = [e]
            if not isiterable(e[0]):
                e = [[el] for el in e]
            # casting dict to qdict based on data on the first row
            for i, val in enumerate(e[0]):
                if type(val) is dict:
                    self.casts[i] = "qdict"
        return e


class TextProcessor(Processor):
    def __init__(self, prs, strings, path=None):
        super().__init__(prs, strings, path)

    # reads a text row as a row with 1 column
    def get_input_iterator(self):
        return ([line.rstrip("\n\r")] for line in self.input_file)


class SpyProcessor(Processor):
    def __init__(self, prs, strings, path=None):
        super().__init__(prs, strings, path)
        self.has_header = True

    def reading_data(self):
        return self.input_col_names

    def handle_header_row(self, row):
        self.input_col_names = row

    @staticmethod
    def unpack_line(line):
        return pickle.loads(bytes.fromhex(line))

    # input is a serialized Python list converted to hex
    def get_input_iterator(self):
        return (self.unpack_line(line[0:-1]) for line in self.input_file)


class JSONProcessor(Processor):
    def __init__(self, prs, strings, path=None, **options):
        import json

        super().__init__(prs, strings, path)
        json.loads('{"a": 1}', **options)  # test options
        self.options = options
        self.input_col_names = ["json"]

        # this might not be the most efficient way of converting None -> NULL, look at:
        # https://stackoverflow.com/questions/27695901/python-jsondecoder-custom-translation-of-null-type
        self.decoder = json.JSONDecoder(
            object_pairs_hook=qdict,
            **options,
        )

    # 1 row = 1 json
    def get_input_iterator(self):
        return ([self.decoder.decode(line)] for line in self.input_file)


class ORJSONProcessor(Processor):
    def __init__(self, prs, strings, path=None, **options):
        super().__init__(prs, strings, path)
        try:
            import orjson
        except ModuleNotFoundError as e:
            # orjson must be installed separately
            spyql.log.user_error(
                "`orjson` module not found. You might need to install it",
                e,
                "pip3 install orjson",
            )
        orjson.loads('{"a": 1}', **options)  # test options (should be empty...)
        self.input_col_names = ["json"]

    # 1 row = 1 json
    def get_input_iterator(self):
        import orjson

        # this might not be the most efficient way of converting None -> NULL, look at:
        # https://stackoverflow.com/questions/27695901/python-jsondecoder-custom-translation-of-null-type
        return ([qdict(orjson.loads(line))] for line in self.input_file)


# CSV
class CSVProcessor(Processor):
    def __init__(
        self,
        prs,
        strings,
        path=None,
        sample_size=10,
        header=None,
        infer_dtypes=True,
        **options,
    ):
        super().__init__(prs, strings, path)
        self.sample_size = sample_size
        self.has_header = header
        self.infer_dtypes = infer_dtypes
        self.options = options
        csv.reader(StringIO("test"), **self.options)  # test options

    def _test_dtype(self, v):
        v = v.strip()
        if not v:
            return (-100, None)  # empty string: do not cast
        try:
            int(v)
            return (10, "int_")
        except ValueError:
            try:
                float(v)
                return (20, "float_")
            except ValueError:
                try:
                    complex(v)
                    return (30, "complex_")
                except ValueError:
                    return (100, None)  # not a basic type: do not cast

    def _infer_dtypes(self, reader):
        if self.has_header:
            next(reader, None)  # skip header
        dtypes_rows = [[self._test_dtype(col) for col in line] for line in reader]
        if dtypes_rows and dtypes_rows[0]:
            dtypes = [
                max([row[c] if c < len(row) else (-100, None) for row in dtypes_rows])
                for c in range(len(dtypes_rows[0]))
            ]
            for c in range(len(dtypes)):
                cast = dtypes[c][1]
                if cast:
                    self.casts[c] = cast

    def get_input_iterator(self):
        # Part 1 reads sample to detect dialect and if has header
        # TODO force linedelimiter to be new line char set

        # saves sample to a string
        # NOTE if dialect is given and type detection is off we should not need a sample
        sample = io.StringIO()
        for line in list(islice(self.input_file, self.sample_size)):
            sample.write(line)
        sample_val = sample.getvalue()

        if not sample_val:
            return []
        if not self.options:
            # CSV dialect and header detection
            try:
                self.options = {"dialect": csv.Sniffer().sniff(sample_val)}
            except Exception as e:
                spyql.log.user_error("Could not detect CSV dialect from input", e)
            if self.has_header is None:
                try:
                    self.has_header = csv.Sniffer().has_header(sample_val)
                except Exception as e:
                    spyql.log.user_error("Could not detect if input CSV has header", e)
        elif self.has_header is None:
            self.has_header = True  # default if dialect is not automatically detected
        sample.seek(0)  # rewinds the sample
        if self.infer_dtypes:
            self._infer_dtypes(csv.reader(sample, **self.options))
            sample.seek(0)  # rewinds the sample again

        return chain(
            csv.reader(
                sample, **self.options
            ),  # goes through sample again (for reading input data)
            csv.reader(self.input_file, **self.options),
        )  # continues to the rest of the file

    def reading_data(self):
        return (not self.has_header) or (self.input_col_names)

    def handle_header_row(self, row):
        self.input_col_names = [make_str_valid_varname(val) for val in row]



"""
.. code-block:: sql

    SELECT *
    FROM json

"""

import operator
from spyql.nulltype import Null
from spyql.qdict import qdict


def _init_aggs():
    """Initializes aggregates tracking mechanism"""
    global _agg_idx
    global _agg_key
    global _aggs
    _agg_idx = 0  # pointer to the current aggregate tracker, reset every new row
    _agg_key = ()  # aggregation key of the current row (identifies the group)
    _aggs = dict()  # cumulative of each aggregation function call


def _start_new_agg_row(key):
    """Resets the aggregates tracking mechanism for starting a new row"""
    global _agg_idx
    global _agg_key
    _agg_idx = 0  # reset aggregate call tracker
    _agg_key = key  # set the group


def _get_aggs():
    global _aggs
    return _aggs


def _agg_op(op, val, default=Null):
    """
    Generic aggregation function.
    `val` is the value for the current aggregation of the current row (ignores NULLs).
    `op` should be `function(cumulative_from_prev_rows, value_for_cur_row)`.
    Current mechanism is based on the order of aggregate function calls in the query.
    This might fail if there are flow control statements, which is not checked by the
    parser! (e.g. `SELECT max_agg(x) if x>0 else 0, count_agg(*)` produces unpredictable
    results because in some rows the max function is executed and in others is not.
    Therefore, in some rows the count_agg will have agg_idx 1 and on others agg_idx 0).
    """
    # TODO prevent/detect erroneous aggregation behavior due to flow control statements
    global _agg_idx
    global _agg_key
    global _aggs
    key = (_agg_key, _agg_idx)
    _agg_idx += 1  # moves to the next aggregation (before any return)
    prev_val = _aggs.get(key, default)
    if val is Null:
        return prev_val
    new_val = val if prev_val is Null else op(prev_val, val)
    _aggs[key] = new_val
    return new_val


# Aggregation functions


def sum_agg(val):
    """Sum across all non-null input values"""
    return _agg_op(operator.add, val)


def prod_agg(val):
    """Product across all non-null input values"""
    return _agg_op(operator.mul, val)


def count_agg(val):
    """Count all non-null input values"""
    """`count_agg(*)` counts the number of input rows"""
    return sum_agg(0 if val is Null else 1)


def avg_agg(val):
    """Average all non-null input values"""
    return sum_agg(val) / count_agg(val)


def min_agg(val):
    """Minimum value across all non-null input values"""
    return _agg_op(min, val)


def max_agg(val):
    """Maximum value across all non-null input values"""
    return _agg_op(max, val)


def list_agg(val, respect_nulls=True):
    """
    Collects all input values into a list.
    Filters out NULLs when `respect_nulls` is `False`.
    """
    vals = sum_agg([val] if respect_nulls or val is not Null else Null)
    return [] if vals is Null else vals  # guarantees that result is always an array


def string_agg(val, sep, respect_nulls=False):
    """
    Concatenates all input values into a string.
    Uses `sep` to separate values in the string.
    Filters out NULLs when `respect_nulls` is `False` (default).
    """
    return str(sep).join(
        sum_agg([str(val)] if respect_nulls or val is not Null else Null)
    )


def set_agg(val, respect_nulls=True):
    """
    Collects all distinct input values into a set.
    Filters out NULLs when `respect_nulls` is `False`.
    """
    return _agg_op(operator.or_, {val} if respect_nulls or val is not Null else set())


def dict_agg(key, val):
    """
    Collects key-value pairs into a dict.
    Key must be unique and not null (null keys are discarded).
    In case of duplicated keys, the value returned is the last seen.
    """
    return _agg_op(
        lambda a_dict, an_entry: a_dict.updatef(an_entry),
        qdict({key: val} if key is not Null else {}),
    )


def first_agg(val, respect_nulls=True):
    """
    Returns the first value.
    Returns the first non-null value when `respect_nulls` is `False`.
    """
    return _agg_op(
        lambda prev, _: prev, [val] if respect_nulls or val is not Null else Null
    )[0]


def last_agg(val, respect_nulls=True):
    """
    Returns the last value.
    Returns the last non-null value when `respect_nulls` is `False`.
    """
    return _agg_op(
        lambda _, cur: cur, [val] if respect_nulls or val is not Null else Null
    )[0]


def lag_agg(val, offset=1, default=Null):
    """
    Returns the value at `offset` rows before the last row. Returns `default` if there
    is no such row.
    Especially useful with `SELECT PARTIAL` to return the value at `offset` rows before
    the current row.
    """
    res = _agg_op(lambda prev, cur: (cur + prev)[: offset + 1], [val], default)
    return res[-1] if len(res) > offset else Null


def count_distinct_agg(val):
    """Count the number of unique (non-null) input values."""
    """`count_distinct_agg(*)` counts the number of distinct rows."""
    return len(_agg_op(operator.or_, Null if val is Null else {val}))


def any_agg(val):
    """Returns True when there is at least one True value, ignoring NULLs"""
    return _agg_op(operator.or_, Null if val is Null else bool(val))


def every_agg(val):
    """Returns True when all non-null values are True"""
    return _agg_op(operator.and_, Null if val is Null else bool(val))



from spyql.nulltype import *
from spyql.sqlfuncs import *
from spyql.qdict import qdict
import numpy as np
import math


def test_arithmetic_operators():
    assert NULL + 2 is NULL
    assert 2 + NULL is NULL
    assert NULL + NULL is NULL
    assert 1 + 2 is not NULL
    assert 2 + 3 + NULL is NULL

    assert NULL - 2 is NULL
    assert 2 - NULL is NULL
    assert NULL - NULL is NULL
    assert 1 - 2 is not NULL
    assert 2 - 3 - NULL is NULL

    assert NULL * 2 is NULL
    assert NULL * 0 is NULL
    assert 2 * NULL is NULL
    assert NULL * NULL is NULL
    assert 1 * 2 is not NULL
    assert 2 * 3 * NULL is NULL

    assert NULL / 2.0 is NULL
    assert NULL / 0.0 is NULL
    assert 2.0 / NULL is NULL
    assert NULL / NULL is NULL
    assert 1.0 / 2.0 is not NULL
    assert 2.0 / 3.0 / NULL is NULL

    assert NULL // 2 is NULL
    assert NULL // 0 is NULL
    assert 2 // NULL is NULL
    assert NULL // NULL is NULL
    assert 1 // 2 is not NULL
    assert 2 // 3 // NULL is NULL

    assert NULL % 2 is NULL
    assert NULL % 0 is NULL
    assert 2 % NULL is NULL
    assert NULL % NULL is NULL
    assert 1 % 2 is not NULL
    assert 2 % 3 % NULL is NULL

    assert NULL ** 2 is NULL
    assert NULL ** 0 is NULL
    assert 2 ** NULL is NULL
    assert NULL ** NULL is NULL
    assert 2 ** 3 is not NULL
    assert 2 ** 3 ** NULL is NULL

    assert -NULL is NULL
    assert +NULL is NULL
    assert abs(NULL) is NULL
    assert round(NULL) is NULL
    assert math.trunc(NULL) is NULL
    assert math.ceil(NULL) is NULL
    assert math.floor(NULL) is NULL


def test_bitwise_operators():
    assert NULL << 2 is NULL
    assert NULL << 0 is NULL
    assert 2 << NULL is NULL
    assert NULL << NULL is NULL
    assert 100 << 3 is not NULL
    assert 100 << 3 << NULL is NULL

    assert NULL >> 2 is NULL
    assert NULL >> 0 is NULL
    assert 2 >> NULL is NULL
    assert NULL >> NULL is NULL
    assert 100 >> 3 is not NULL
    assert 100 >> 3 >> NULL is NULL

    assert NULL & 2 is NULL
    assert NULL & 0 is NULL
    assert 2 & NULL is NULL
    assert NULL & NULL is NULL
    assert 100 & 3 is not NULL
    assert 100 & 3 & NULL is NULL

    assert NULL | 2 is NULL
    assert NULL | 0 is NULL
    assert 2 | NULL is NULL
    assert NULL | NULL is NULL
    assert 100 | 3 is not NULL
    assert 100 | 3 | NULL is NULL

    assert NULL ^ 2 is NULL
    assert NULL ^ 0 is NULL
    assert 2 ^ NULL is NULL
    assert NULL ^ NULL is NULL
    assert 100 ^ 3 is not NULL
    assert 100 ^ 3 ^ NULL is NULL

    assert ~NULL is NULL
    assert ~~NULL is NULL


def test_bool_operators():
    assert not NULL  # NULL bool value is False

    assert (NULL and True) is NULL
    assert (True and NULL) is NULL

    assert (NULL and False) is NULL
    assert (
        False and NULL
    ) is False  # inconsistent behaviour... (despite NULL evals to False)

    assert (NULL and NULL) is NULL

    assert (NULL or True) is True
    assert (True or NULL) is True

    assert (NULL or False) is False
    assert (
        False or NULL
    ) is NULL  # inconsistent behaviour... (despite NULL evals to False)

    assert (NULL or NULL) is NULL


def test_comparison_operators():
    assert (NULL == 2) is NULL
    assert (2 == NULL) is NULL
    assert (NULL == NULL) is NULL

    assert (NULL != 2) is NULL
    assert (2 != NULL) is NULL
    assert (NULL != NULL) is NULL

    assert (NULL > 2) is NULL
    assert (2 > NULL) is NULL
    assert (NULL > NULL) is NULL

    assert (NULL >= 2) is NULL
    assert (2 >= NULL) is NULL
    assert (NULL >= NULL) is NULL

    assert (NULL < 2) is NULL
    assert (2 < NULL) is NULL
    assert (NULL < NULL) is NULL

    assert (NULL <= 2) is NULL
    assert (2 <= NULL) is NULL
    assert (NULL <= NULL) is NULL


def test_container_operators():
    assert len(NULL) == 0
    assert len([x for x in NULL]) == 0
    assert len([x for x in reversed(NULL)]) == 0
    assert NULL["abc"] is NULL
    assert NULL[1] is NULL
    assert NULL.get("abc") is NULL
    assert NULL.get("abc", None) is NULL
    assert not (1 in NULL)
    assert not ("a" in NULL)
    assert not (NULL in NULL)
    assert NULL in [NULL]


def test_matrix_mult():
    assert NULL @ np.array([1, 2]) is NULL
    assert NULL @ np.array([]) is NULL
    assert np.array([1, 2]) @ NULL is NULL
    assert NULL @ NULL is NULL
    assert np.array([1, 2]) @ np.array([3, 4]) is not NULL
    assert np.array([1, 2]) @ np.array([3, 4]) @ NULL is NULL


def test_casting():
    assert float_("1.1") == float("1.1")
    assert float_("-1.21e-5") == float("-1.21e-5")
    assert float_(1) == float(1)
    assert float_("") is NULL
    assert float_("abc") is NULL
    assert float_(NULL) is NULL

    assert int_("1") == int("1")
    assert int_("-981") == int("-981")
    assert int_(1) == int(1)
    assert int_("") is NULL
    assert int_("abc") is NULL
    assert int_(NULL) is NULL

    assert str_("1") == str("1")
    assert str_(1) == str(1)
    assert str_(-981) == str(-981)
    assert str_(-1.21e-5) == str(-1.21e-5)
    assert str_(NULL) is NULL

    assert complex_("1.1") == complex("1.1")
    assert complex_("1.1+2j") == complex("1.1+2j")
    assert complex_(1) == complex(1)
    assert complex_("") is NULL
    assert complex_("abc") is NULL
    assert complex_(NULL) is NULL


def test_aux_functions():
    assert coalesce(1, 2) == 1
    assert coalesce(NULL, 2) == 2
    assert ifnull(1, 2) == 1
    assert ifnull(NULL, 2) == 2
    assert nullif(1, 1) is NULL
    assert nullif(1, 2) == 1


def test_misc():
    assert null is NULL
    assert Null is NULL


def test_dict():
    qdict({})["abc"] is NULL
    qdict({"abc": None})["abc"] is NULL
    qdict({"abc": None})["abc"] is not None  # Attention!
    qdict({"abc": 1})["abc"] is not NULL
    qdict({"abc": "def"})["abc"] is not NULL
    qdict({})["abc"]["def"]["hij"] is NULL
    qdict({"abc": {"def": 1}})["abc"]["def"] is not NULL



from spyql.nulltype import Null


class OutputHandler:
    """Mediates data processing with data writting"""

    @staticmethod
    def make_handler(prs):
        """
        Chooses the right handler depending on the kind of query
        and eventual optimization opportunities
        """
        if prs["group by"] and not prs["partials"]:
            return GroupByDelayedOutSortAtEnd(
                prs["order by"], prs["limit"], prs["offset"]
            )
        if prs["order by"]:
            # TODO optimization: use special handler that only keeps the top n elements
            #   in memory when LIMIT is defined
            if prs["distinct"]:
                return DistinctDelayedOutSortAtEnd(
                    prs["order by"], prs["limit"], prs["offset"]
                )
            return DelayedOutSortAtEnd(prs["order by"], prs["limit"], prs["offset"])
        if prs["distinct"]:
            return LineInDistinctLineOut(prs["limit"], prs["offset"])
        return LineInLineOut(prs["limit"], prs["offset"])

    def __init__(self, limit, offset):
        self.limit = limit
        self.rows_written = 0
        self.offset = offset if offset else 0

    def set_writer(self, writer):
        self.writer = writer

    def handle_result(self, result, group_key, sort_keys):
        """
        To be implemented by child classes to handle a new output row (aka result).
        All inputs should be tuples.
        """
        return self.is_done()

    def is_done(self):
        # premature ending
        return self.limit is not None and self.rows_written >= self.limit

    def write(self, row):
        if self.offset > 0:
            self.offset = self.offset - 1
        else:
            self.writer.writerow(row)
            self.rows_written = self.rows_written + 1

    def finish(self):
        self.writer.flush()


class LineInLineOut(OutputHandler):
    """Simple handler that immediately writes every processed row"""

    def handle_result(self, result, *_):
        self.write(result)
        return self.is_done()

    def finish(self):
        super().finish()


class LineInDistinctLineOut(OutputHandler):
    """In-memory distinct handler that immediately writes every non-duplicated row"""

    def __init__(self, limit, offset):
        super().__init__(limit, offset)
        self.output_rows = set()

    def handle_result(self, result, *_):
        # uses a dict to store distinct results instead of storing all rows
        if result in self.output_rows:
            return False  # duplicate

        self.output_rows.add(result)
        self.write(result)
        return self.is_done()

    def finish(self):
        super().finish()


class DelayedOutSortAtEnd(OutputHandler):
    """
    Only writes after collecting and sorting all data.
    Temporary implementation that reads every processed row into memory.
    """

    def __init__(self, orderby, limit, offset):
        super().__init__(limit, offset)
        self.orderby = orderby
        self.output_rows = []

    def handle_result(self, result, sort_keys, *_):
        self.output_rows.append({"data": result, "sort_keys": sort_keys})
        # TODO use temporary files to write `output_rows` whenever it gets too large
        # TODO sort intermediate results before writing to a temporary file
        return False  # no premature endings here

    def finish(self):
        # TODO read and merge previously sorted temporary files (look into heapq.merge)
        # 1. sorts everything
        if self.orderby:
            for i in reversed(range(len(self.orderby))):
                # taking advantage of list.sort being stable to sort elements from minor
                # to major criteria (not be the most efficient way but straightforward)
                self.output_rows.sort(
                    key=lambda row: (
                        # handle of NULLs based on NULLS FIRST/LAST specification
                        (row["sort_keys"][i] is Null) != self.orderby[i]["rev_nulls"],
                        row["sort_keys"][i],
                    ),
                    reverse=self.orderby[i]["rev"],  # handles ASC/DESC order
                )
        # 2. writes sorted rows to output
        for row in self.output_rows:
            # it would be more efficient to slice `output_rows` based on limit/offset
            # however, this is more generic with less repeated logic and this is a
            # temporary implementation
            if self.is_done():
                break
            self.write(row["data"])
        super().finish()


class GroupByDelayedOutSortAtEnd(DelayedOutSortAtEnd):
    """
    Extends `DelayedOutSortAtEnd` to only store intermediate group by results instead of
    keeping all rows in memory
    """

    def __init__(self, orderby, limit, offset):
        super().__init__(orderby, limit, offset)
        self.output_rows = dict()

    def handle_result(self, result, sort_keys, group_key):
        # uses a dict to store intermidiate group by results instead of storing all rows
        self.output_rows[group_key] = {"data": result, "sort_keys": sort_keys}
        return False  # no premature endings here

    def finish(self):
        #  converts output_rows dict to list so that it can be sorted and written
        self.output_rows = list(self.output_rows.values())
        super().finish()


class DistinctDelayedOutSortAtEnd(DelayedOutSortAtEnd):
    """
    Alters `DelayedOutSortAtEnd` to only store distinct results instead of
    keeping all rows in memory
    """

    def __init__(self, orderby, limit, offset):
        super().__init__(orderby, limit, offset)
        self.output_rows = dict()

    def handle_result(self, result, sort_keys, *_):
        # uses a dict to store distinct results instead of storing all rows
        if result not in self.output_rows:
            self.output_rows[result] = sort_keys
        return False  # no premature endings here

    def finish(self):
        # converts output_rows dict to list so that it can be sorted and written
        self.output_rows = [
            {"data": k, "sort_keys": v} for k, v in self.output_rows.items()
        ]
        super().finish()



import cProfile
import spyql.cli

"""
Run `python3 -m spyql.prof <my_query>` for profiling. 
Profiler stats saved on file `spyql.stats`. 
"""

if __name__ == "__main__":
    cProfile.run("spyql.cli.main()", "spyql.stats")
